#include "collection.h"

constexpr std::string_view downloads("downloads");

static const std::regex remote("(^https?):\\/\\/");

Collection::Collection(
    const std::filesystem::path& location,
    const std::filesystem::path& destination,
    const std::string& target,
    const std::shared_ptr<Logger>& logger,
    const std::shared_ptr<TaskManager>& taskManager
):
    Loggable(logger),
    mutex(),
    destination(destination),
    target(target),
    taskManager(taskManager),
    root(std::make_unique<Component>(location, this, logger)),
    librariesPath(),
    knownSources(),
    successSources(),
    pendingSources(),
    downloadingSources(),
    readingSources(),
    buildingSources()
{
    try {
        std::filesystem::create_directories(destination);
    } catch (const std::exception& e) {
        fatal(e.what());

        throw e;
    }

    root->read();
    if (!root->successfullyRead())
        throw std::runtime_error("Error reading project in " + location.string());

    if (root->getType() != Component::mason)
        throw std::runtime_error("Project in " + location.string() + " doesn't seem to be a valid mason project");

    librariesPath = root->getLibrariesPath();
    std::lock_guard lock(mutex);
    while (pendingSources.size() > 0) {
        Source source = pendingSources.front();
        pendingSources.pop();
        processSource(source);
    }

    taskManager->queue(
        std::bind(&Component::build, root.get(), destination, target, false),
        std::bind(&Collection::rootBuilt, this, std::placeholders::_1)
    );
}

void Collection::addSource(const Source& source) {
    std::lock_guard lock(mutex);
    std::pair<Sources::const_iterator, bool> result = knownSources.emplace(source);
    if (!result.second) {
        major("Source " + source.print() + " is already present, skipping as duplicate");
        return;
    }

    debug("Adding source " + source.print());
    switch (root->getState()) {
        case Component::initial:
        case Component::reading:
            pendingSources.emplace(source);
            break;
        case Component::ready:
        case Component::building:
        case Component::done:
            processSource(source);
            break;
        case Component::error:
            throw std::runtime_error("Project in " + root->getLocation().string() + " doesn't seem to be a valid mason project");
    }
}

void Collection::processSource(const Source& source) {
    std::string path = source.getPath();
    if (isRemote(path))
        queueDownload(source);
    else
        queueRead(source, path);
}

void Collection::sourceDownaloded(const Source& source, TaskManager::Error err) {
    std::lock_guard lock(mutex);

    Downloads::const_iterator itr = downloadingSources.find(source);
    std::optional<std::filesystem::path> location = itr->second->getLocation();
    downloadingSources.erase(itr);
    if (err.has_value() || !location.has_value()) {
        error("Coundn't download " + source.print() + ": " + err.value()->what());
        return;
    }

    queueRead(source, location.value());
}

void Collection::sourceRead(const Source& source, TaskManager::Error err) {
    std::lock_guard lock(mutex);

    Components::iterator itr = readingSources.find(source);
    if (err.has_value()) {
        error("Coundn't read " + source.print() + ": " + err.value()->what());
        readingSources.erase(itr);
        return;
    }

    if (!itr->second->successfullyRead()) {
        error("Source " + source.print() + " had an error during read process, canceling build of this component");
        readingSources.erase(itr);
        return;
    }

    debug("Source " + source.print() + " has been read, scheduling build");
    std::pair<Components::iterator, bool> res = buildingSources.emplace(source, std::move(itr->second));
    readingSources.erase(itr);

    std::string tgt = target;
    if (source.hasExplicitTarget())
        tgt = source.getTarget();

    taskManager->queue(
        std::bind(&Component::build, res.first->second.get(), destination/librariesPath, tgt, true),
        std::bind(&Collection::sourceBuilt, this, source, std::placeholders::_1)
    );
}

void Collection::sourceBuilt(const Source& source, TaskManager::Error err) {
    std::lock_guard lock(mutex);
    Components::iterator itr = buildingSources.find(source);
    bool success = itr->second->successfullyBuilt();
    buildingSources.erase(itr);
    if (err.has_value()) {
        error("Coundn't read " + source.print() + ": " + err.value()->what());
        return;
    }

    if (!success) {
        error("Source " + source.print() + " had an error during build process");
        return;
    }

    debug("Source " + source.print() + " has been built");
    successSources.emplace(source);
}

void Collection::rootBuilt(TaskManager::Error err) {
    std::string loc = root->getLocation();
    if (err.has_value()) {
        error("Coundn't build project at" + loc + ": " + err.value()->what());
        return;
    }

    if (root->successfullyBuilt())
        info("Project at " + loc + " was successfully built");
    else
        error("Project at " + loc + " had some errors during build process");
}

bool Collection::isRemote(const std::string& source) {
    return std::regex_search(source, remote);
}

void Collection::queueDownload(const Source& source) {
    debug("Scheduling download of " + source.print());
    std::pair<Downloads::iterator, bool> res = downloadingSources.emplace(std::piecewise_construct,
        std::forward_as_tuple(source),
        std::forward_as_tuple(
            std::make_unique<Download>(source.getPath(), destination/downloads, logger)
        )
    );

    taskManager->queue(
        std::bind(&Download::proceed, res.first->second.get()),
        std::bind(&Collection::sourceDownaloded, this, source, std::placeholders::_1)
    );
}

void Collection::queueRead(const Source& source, const std::filesystem::path& location) {
    debug("Scheduling read of " + source.print());
    std::pair<Components::iterator, bool> res = readingSources.emplace(std::piecewise_construct,
        std::forward_as_tuple(source),
        std::forward_as_tuple(
            std::make_unique<Component>(location, this, logger, source.getOptionalName())
        )
    );

    taskManager->queue(
        std::bind(&Component::read, res.first->second.get()),
        std::bind(&Collection::sourceRead, this, source, std::placeholders::_1)
    );
}

bool Collection::hasSource(const Source& source) const {
    std::lock_guard lock(mutex);
    return _hasSource(source);
}

unsigned int Collection::sourcesTotal() const {
    std::lock_guard lock(mutex);
    return _sourcesTotal();
}

unsigned int Collection::sourcesSuccess() const {
    std::lock_guard lock(mutex);
    return _sourcesSuccess();
}

unsigned int Collection::sourcesPending() const {
    std::lock_guard lock(mutex);
    return _sourcesPending();
}

unsigned int Collection::sourcesError() const {
    std::lock_guard lock(mutex);
    return _sourcesError();
}

bool Collection::success() const {
    std::lock_guard lock(mutex);
    return _sourcesTotal() == _sourcesSuccess() && root->successfullyBuilt();
}

bool Collection::_hasSource(const Source& source) const {
    return knownSources.count(source) > 0;
}

unsigned int Collection::_sourcesTotal() const {
    return knownSources.size();
}

unsigned int Collection::_sourcesSuccess() const {
    return successSources.size();
}

unsigned int Collection::_sourcesPending() const {
    return downloadingSources.size() + readingSources.size() + buildingSources.size() + pendingSources.size();
}

unsigned int Collection::_sourcesError() const {
    return _sourcesTotal() - _sourcesPending() - _sourcesSuccess();
}
